﻿using System;
using System.Windows.Input;
using Microsoft.Win32;
using WpfAudioPlayer.Commands;
using WpfAudioPlayer.Models;
using WpfAudioPlayer.Views;
using System.Windows;

namespace WpfAudioPlayer.ViewModels
{
	/// <summary>
	/// メイン ウィンドウの View と Model を仲介するクラスです。
	/// </summary>
	class MainWindowViewModel : ViewModelBase, IDisposable
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="owner">オーナーとなるウィンドウ。</param>
		public MainWindowViewModel( Window owner )
		{
			this._owner = owner;

			this.PlayerViewModel = new PlayerViewModel();
			this.PlayerViewModel.RequestNextSong += ( sender, e ) =>
			{
				var song = this.SongListViewModel.GetNextSong( e.IsNext, e.Song );
				if( song != null )
				{
					this.SongListViewModel.SelectedSong = song;
				}

				return song;
			};

			this.SongListViewModel = new SongListViewModel();
			this.SongListViewModel.SongSelecttionChanged += ( sender, e ) =>
			{
				this.PlayerViewModel.CurrentSong = e.Song;
			};

			this.InitializeCommand();
		}
		
		/// <summary>
		/// コマンドを初期化します。
		/// </summary>
		private void InitializeCommand()
		{
			// 再生
			this.ImportContentCommand = new DelegateCommand(
				() =>
				{
					var dialog = new OpenFileDialog() { Filter = this.PlayerViewModel.FileFilter, Multiselect = true };
					var isOK   = dialog.ShowDialog();

					if( isOK.Value )
					{
						this.SongListViewModel.Add( dialog.FileNames );
					}
				}
				);

			// イコライザーを表示
			this.ShowEqualizer = new DelegateCommand(
				() =>
				{
					var window = new EqualizerWindow() { Owner = this._owner, Title = "イコライザー" };
					window.DataContext = this.PlayerViewModel.EqualizerViewModel;

					window.Closed += ( sender, e ) =>
					{
						this._canShowEqualizer = true;
						this.NotifyPropertyChanged( "ShowEqualizer" );
					};

					this._canShowEqualizer = false;
					this.NotifyPropertyChanged( "ShowEqualizer" );
					window.Show();
				},
				() =>
				{
					return this._canShowEqualizer;
				}
				);
		}

		/// <summary>
		/// PlayerViewModel を取得します。
		/// </summary>
		public PlayerViewModel PlayerViewModel { get; private set; }

		/// <summary>
		/// SongListViewModel を取得します。
		/// </summary>
		public SongListViewModel SongListViewModel { get; private set; }

		/// <summary>
		/// アプリケーションの状態を示すテキストを取得します。
		/// </summary>
		public string Status
		{
			get
			{
				return this._status;
			}
			private set
			{
				this._status = value;
				this.NotifyPropertyChanged( "Status" );
			}
		}

		/// <summary>
		/// 音楽ファイルを取り込みます。
		/// </summary>
		public ICommand ImportContentCommand { get; private set; }

		/// <summary>
		/// イコライザーを表示します。
		/// </summary>
		public ICommand ShowEqualizer { get; private set; }

		#region フィールド

		/// <summary>
		/// アプリケーションの状態を示すテキスト。
		/// </summary>
		private string _status;

		/// <summary>
		/// イコライザーの表示が可能である事を示す値。
		/// </summary>
		private bool _canShowEqualizer = true;

		/// <summary>
		/// オーナーとなるウィンドウ。
		/// </summary>
		private Window _owner;

		#endregion

		#region IDisposable メンバ

		/// <summary>
		/// リソースの解放を行います。
		/// </summary>
		public void Dispose()
		{
			if( this.PlayerViewModel != null )
			{
				this.PlayerViewModel.Dispose();
				this.PlayerViewModel = null;
			}

			if( this.SongListViewModel != null )
			{
				this.SongListViewModel = null;
			}
		}

		#endregion
	}
}
