﻿using System;
using System.Globalization;
using System.Windows.Input;
using System.Windows.Threading;
using WpfAudioPlayer.Commands;
using WpfAudioPlayer.Models;
using System.Collections.Generic;

namespace WpfAudioPlayer.ViewModels
{
	/// <summary>
	/// プレイヤーの View と Model を仲介するクラスです。
	/// </summary>
	sealed class PlayerViewModel : ViewModelBase, IDisposable
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <exception cref="Exception">初期化に失敗した。</exception>
		public PlayerViewModel()
		{
			this._timer           = new DispatcherTimer();
			this._timer.Tick     += new EventHandler( OnTick );
			this._timer.Interval  = TimeSpan.FromSeconds( 1.0 );
			this.Volume           = 1.0f;

			// BASS の初期化
			{
				var folderPath = AppDomain.CurrentDomain.BaseDirectory + "BASS";
				AudioPlayer.BassInitialize( folderPath );

				this.FileFilter = WpfAudioPlayer.Models.AudioPlayer.FileFilter;
			}

			// 再生状態
			this._statusTexts = new Dictionary< PlayState, string >()
			{
				{ Models.PlayState.Paused,  "一時停止" },
				{ Models.PlayState.Playing, "再生中"   },
				{ Models.PlayState.Stopped, "停止"     }
			};

			this.EqualizerViewModel = new EqualizerViewModel();

			this.InitializeCommand();
		}
	
		/// <summary>
		/// 次に再生対象とする曲を取得する。
		/// </summary>
		/// <param name="isNext">次の曲を要求する場合は true、前の曲ならば false。</param>
		/// <returns>次に再生対象とする曲。</returns>
		private SongViewModel GetNextSong( bool isNext )
		{
			if( this.RequestNextSong == null ) { return null; }

			return this.RequestNextSong( this, new RequestNextSongEventArgs( isNext, this.CurrentSong ) );
		}

		/// <summary>
		/// コマンドを初期化します。
		/// </summary>
		private void InitializeCommand()
		{
			// 再生
			this.PlayCommand = new DelegateCommand(
				() =>
				{
					this.PlayPause();
				},
				() =>
				{
					return ( this._currentSong != null );
				}
				);

			// 停止
			this.StopCommand = new DelegateCommand(
				() =>
				{
					this.Stop();
				},
				() =>
				{
					return ( this._player != null && this._player.PlayState != PlayState.Stopped );
				}
				);

			// 前へ
			this.BackCommand = new DelegateCommand(
				() =>
				{
					this.MoveSong( false );
				},
				() =>
				{
					return ( this.CurrentSong != null );
				}
				);

			// 次へ
			this.NextCommand = new DelegateCommand(
				() =>
				{
					this.MoveSong( true );
				},
				() =>
				{
					return ( this.CurrentSong != null );
				}
				);
		}

		/// <summary>
		/// 次に再生対象とする曲を選択します。
		/// </summary>
		/// <param name="isNext">次の曲を要求する場合は true、前の曲ならば false。</param>
		/// <returns>次に再生対象とする曲。</returns>
		private void MoveSong( bool isNext )
		{
			var song = this.GetNextSong( isNext );
			if( song != null )
			{
				if( this.PlayState == PlayState.Stopped )
				{
					this.CurrentSong = song;
				}
				else
				{
					this.Stop();
					this.CurrentSong = song;
					this.PlayPause();
				}
			}
		}

		/// <summary>
		/// 音声の再生中に呼び出されるタイマー イベント。
		/// </summary>
		/// <param name="sender">イベント送信元。</param>
		/// <param name="e">イベントデータ。</param>
		private void OnTick( object sender, EventArgs e )
		{
			if( this._player == null || this._player.PlayState != PlayState.Playing ) { return; }

			if( this._player.CurrentTime >= this._player.Duration )
			{
				this.Stop();
				this.MoveSong( true );
				this.PlayPause();
			}
			else
			{
				this._currentTime = PlayerViewModel.TimeToText( this._player.CurrentTime );
				this.NotifyPropertyChanged( "Position", "CurrentTime" );
			}
		}

		/// <summary>
		/// コンテンツの再生・一時停止を行います。
		/// </summary>
		private void PlayPause()
		{
			if( this.PlayState == PlayState.Playing )
			{
				this._player.Pause();
				this.IsPlaying = false;
				this.NotifyPropertyChanged( "IsPlaying", "Status" );
			}
			else
			{
				if( this._player == null )
				{
					this._player = new AudioPlayer( this._currentSong.FilePath );
					this._player.Volume = this._volume;
					this.NotifyPropertyChanged( "Length" );

					this.EqualizerViewModel.Attach( this._player.Stream );
				}

				this._player.Play();
				this._timer.Start();
				this.IsPlaying = true;
				this.NotifyPropertyChanged( "CanSeek", "IsPlaying", "Status" );
			}
		}

		/// <summary>
		/// コンテンツの再生を停止します。
		/// </summary>
		private void Stop()
		{
			if( this._player != null )
			{
				this._timer.Stop();
				this._player.Stop();
			}

			this._currentTime = PlayerViewModel.EmptyTimeText;
			this.IsPlaying = false;
			this.NotifyPropertyChanged( "CurrentTime", "CanSeek", "Position", "IsPlaying", "StopCommand", "Status" );
		}

		/// <summary>
		/// 時間を文字列に変換します。
		/// </summary>
		/// <param name="time">時間。</param>
		/// <returns>変換された文字列。</returns>
		private static string TimeToText( TimeSpan time )
		{
			return String.Format( CultureInfo.InvariantCulture, "{0:d2}:{1:d2}:{2:d2}", time.Hours, time.Minutes, time.Seconds );
		}

		#region イベント

		/// <summary>
		/// 次に再生対象とする曲を要求するイベントハンドラ。
		/// </summary>
		/// <param name="sender">イベント送信元。</param>
		/// <param name="e">イベントデータ。</param>
		/// <returns>次に再生対象とする曲。</returns>
		public delegate SongViewModel RequestNextSongEventHandler( object sender, RequestNextSongEventArgs e );

		/// <summary>
		/// 次に再生すべき曲を要求するイベント。
		/// </summary>
		public event RequestNextSongEventHandler RequestNextSong;

		#endregion

		#region プロパティ

		/// <summary>
		/// コンテンツの再生位置が移動可能である事を示す値を取得します。
		/// </summary>
		public bool CanSeek
		{
			get
			{
				return ( this._player == null ? false : ( this._player.PlayState != PlayState.Stopped && this._player.CanSeek ) );
			}
		}

		/// <summary>
		/// 現在選択されている曲情報を取得または設定します。
		/// </summary>
		public SongViewModel CurrentSong
		{
			get
			{
				return this._currentSong;
			}
			set
			{
				if( this.PlayState != PlayState.Stopped ) { return; }

				this._currentSong = value;
				if( this._currentSong == null ) { return; }

				if( this._player != null )
				{
					this._player.Dispose();
					this._player = null;
				}

				this._duration = value.Duration;

				this.Title = String.Format( CultureInfo.InvariantCulture, "{0} / {1}", this._currentSong.Title, this._currentSong.Artist );
				this.NotifyPropertyChanged( "Title", "Duration" );
			}
		}

		/// <summary>
		/// コンテンツの現在の再生時間を表す文字列を取得します。
		/// </summary>
		public string CurrentTime
		{
			get
			{
				return this._currentTime;
			}
		}
		
		/// <summary>
		/// EqualizerViewModel を取得します。
		/// </summary>
		public EqualizerViewModel EqualizerViewModel { get; set; }

		/// <summary>
		/// ファイル ダイアログに指定する為のフィルタ文字列を取得します。
		/// </summary>
		public string FileFilter { get; private set; }

		/// <summary>
		/// コンテンツの総再生時間を表す文字列を取得します。
		/// </summary>
		public string Duration
		{
			get
			{
				return this._duration;
			}
		}

		/// <summary>
		/// 再生状態である事を示す値を取得または設定します。
		/// </summary>
		public bool IsPlaying { get; set; }

		/// <summary>
		/// コンテンツの総再生時間 ( 秒単位 ) を取得または設定します。
		/// </summary>
		public double Length
		{
			get
			{
				return ( this._player == null ? double.MaxValue : this._player.Duration.TotalSeconds );
			}
		}

		/// <summary>
		/// 音楽再生の状態を取得します。
		/// </summary>
		public PlayState PlayState
		{
			get
			{
				return ( this._player == null ? PlayState.Stopped : this._player.PlayState );
			}
		}

		/// <summary>
		/// プレイヤーの状態を取得します。
		/// </summary>
		public string Status
		{
			get
			{
				return ( this._player == null ? this._statusTexts[ Models.PlayState.Stopped ] : this._statusTexts[ this._player.PlayState ] );
			}
		}

		/// <summary>
		/// コンテンツの再生位置 ( 秒単位 ) を取得または設定します。
		/// </summary>
		public double Position
		{
			get
			{
				return ( this._player == null ? 0.0 : this._player.CurrentTime.TotalSeconds );
			}
			set
			{
				if( this._player != null )
				{
					this._player.CurrentTime = TimeSpan.FromSeconds( value );
					this._currentTime = PlayerViewModel.TimeToText( this._player.CurrentTime );
					this.NotifyPropertyChanged( "Position", "CurrentTime" );
				}
			}
		}

		/// <summary>
		/// コンテンツの題名を取得します。
		/// </summary>
		public string Title { get; private set; }

		/// <summary>
		/// コンテンツ再生時の音量を取得または設定します。
		/// </summary>
		public float Volume
		{
			get
			{
				return this._volume;
			}
			set
			{
				this._volume = value;
				if( this._player != null )
				{
					this._player.Volume = value;
				}

				this._volumeText = String.Format( CultureInfo.InvariantCulture, "音量 : {0}%", ( int )( this._volume * 100 ) );
				this.NotifyPropertyChanged( "Volume", "VolumeText" );
			}
		}

		/// <summary>
		/// コンテンツ再生時の音量を表すテキストを取得します。
		/// </summary>
		public string VolumeText
		{
			get
			{
				return this._volumeText;
			}
		}

		#endregion

		#region コマンド

		/// <summary>
		/// 前のコンテンツを選択します。コンテンツが再生中の場合は、新たに選択されたコンテンツの再生に切り替えます。
		/// </summary>
		public ICommand BackCommand { get; private set; }

		/// <summary>
		/// 次のコンテンツを選択します。コンテンツが再生中の場合は、新たに選択されたコンテンツの再生に切り替えます。
		/// </summary>
		public ICommand NextCommand { get; private set; }

		/// <summary>
		/// コンテンツの再生を開始します。
		/// </summary>
		public ICommand PlayCommand { get; private set; }

		/// <summary>
		/// コンテンツの再生を停止します。
		/// </summary>
		public ICommand StopCommand { get; private set; }

		#endregion

		#region フィールド

		/// <summary>
		/// 無効な時間情報を表す文字列。
		/// </summary>
		private const string EmptyTimeText = "00:00:00";

		/// <summary>
		/// 現在選択されている曲情報。
		/// </summary>
		private SongViewModel _currentSong;

		/// <summary>
		/// コンテンツの現在の再生時間を表す文字列。
		/// </summary>
		private string _currentTime = PlayerViewModel.EmptyTimeText;

		/// <summary>
		/// コンテンツの現在の再生時間を表す文字列。
		/// </summary>
		private string _duration = PlayerViewModel.EmptyTimeText;

		/// <summary>
		/// オーディオ再生を行う為のオブジェクト。
		/// </summary>
		private AudioPlayer _player;

		/// <summary>
		/// オーディオの再生状態を示すテキストのディクショナリ。
		/// </summary>
		private Dictionary< PlayState, string > _statusTexts;

		/// <summary>
		/// 再生状態の定期監視を行う為のタイマー。
		/// </summary>
		private DispatcherTimer _timer;

		/// <summary>
		/// コンテンツ再生時の音量。
		/// </summary>
		private float _volume;

		/// <summary>
		/// コンテンツ再生時の音量を示すテキスト。
		/// </summary>
		private string _volumeText;

		#endregion

		#region IDisposable メンバ

		/// <summary>
		/// リソースの解放を行います。
		/// </summary>
		public void Dispose()
		{
			this.Stop();

			if( this._player != null )
			{
				this._player.Dispose();
				this._player = null;
			}

			AudioPlayer.BassFree();
		}

		#endregion
	}
	
	/// <summary>
	/// 次に再生対象とする曲を要求するイベントのデータ。
	/// </summary>
	class RequestNextSongEventArgs : EventArgs
	{
		/// <summary>
		/// インスタンスを初期化する。
		/// </summary>
		/// <param name="isNext">次の曲を要求する場合は true、前の曲ならば false。</param>
		/// <param name="isNext">現在、再生対象となっている曲。</param>
		public RequestNextSongEventArgs( bool isNext, SongViewModel song )
		{
			this.IsNext = isNext;
			this.Song = song;
		}

		/// <summary>
		/// 次の曲を要求する事を示す値を取得する。
		/// </summary>
		public bool IsNext { get; private set; }

		/// <summary>
		/// 現在、再生対象となっている曲を取得する。
		/// </summary>
		public SongViewModel Song { get; private set; }
	}
}
